import pytest

from kim.field import Field, FieldInvalid, FieldError
from kim.pipelines.base import (
    Session,
    get_data_from_source, get_data_from_name, update_output_to_name,
    update_output_to_source, set_default)


def test_get_data_from_name_pipe():

    data = {
        'name': 'mike',
        'test': 'true',
        'falsy': 0,
        'nested': {'foo': 'bar'}
    }

    default = 'bar'
    field = Field(name='foo', required=True)
    field2 = Field(name='foo', default=default, required=False)
    field3 = Field(name='foo', allow_none=False, required=False)
    field4 = Field(name='falsy', required=True)
    field5 = Field(name='falsy', allow_none=False)

    # Required but not present and no default
    output = {}
    session = Session(field, data, output)
    with pytest.raises(FieldInvalid):
        get_data_from_name(session)

    # Not present but default set
    session = Session(field2, data, output)
    assert get_data_from_name(session) == default

    # Not present and none not allowed
    session = Session(field3, data, output)
    with pytest.raises(FieldInvalid):
        get_data_from_name(session)

    # Required, value present and falsy - should still be allowed
    session = Session(field4, data, output)
    assert get_data_from_name(session) == 0

    # None not allowed, value present and falsy - should still be allowed
    session = Session(field5, data, output)
    assert get_data_from_name(session) == 0


def test_get_data_from_source_pipe():
    data = {
        'name': 'mike'
    }
    output = {}

    field = Field(source='foo')
    session = Session(field, data, output)
    assert get_data_from_source(session) is None

    field = Field(source='name')
    session = Session(field, data, output)
    assert get_data_from_source(session) == 'mike'


def test_get_data_from_source_pipe_dot_syntax():
    data = {
        'user': {
            'name': 'mike'
        }
    }
    output = {}

    field = Field(source='foo.bar')
    session = Session(field, data, output)
    assert get_data_from_source(session) is None

    field = Field(source='user.name')
    session = Session(field, data, output)
    assert get_data_from_source(session) == 'mike'


def test_get_data_from_source_pipe_self():
    data = {
        'name': 'mike'
    }
    output = {}

    field = Field(source='__self__')
    session = Session(field, data, output)
    assert get_data_from_source(session) is data


def test_update_output_to_name_with_dict():

    data = {
        'name': 'mike',
        'test': 'true',
        'nested': {'foo': 'bar'}
    }

    output = {}

    field = Field(name='name', required=True)
    session = Session(field, data, output)
    session.data = data['name']
    update_output_to_name(session)
    assert output == {'name': 'mike'}


def test_update_output_to_name_with_escape():

    data = {
        "foo.bar": "mike",
    }

    output = {}

    field = Field(name="foo\\.bar", required=True)
    session = Session(field, data, output)
    session.data = data["foo.bar"]
    update_output_to_name(session)
    assert output == {"foo.bar": "mike"}


def test_update_output_to_source_with_object():

    data = {
        'source': 'mike',
        'test': 'true',
        'nested': {'foo': 'bar'}
    }

    class MyObject(object):
        pass

    output = MyObject()
    field = Field(source='source', required=True)
    session = Session(field, data, output)
    session.data = data['source']

    update_output_to_source(session)
    assert output.source == 'mike'


def test_update_output_to_source_with_object_dot_notiation():

    class MyObject(object):
        pass

    output = MyObject()
    output.nested = MyObject()

    field = Field(source='nested.source', required=True)
    session = Session(field, {'name': 'mike'}, output)
    session.data = 'mike'

    update_output_to_source(session)
    assert output.nested.source == 'mike'


def test_update_output_to_source_with_dict():

    data = {
        'source': 'mike',
        'test': 'true',
        'nested': {'foo': 'bar'}
    }

    output = {}

    field = Field(source='source', required=True)
    session = Session(field, data, output)
    session.data = data['source']
    update_output_to_source(session)
    assert output == {'source': 'mike'}


def test_update_output_to_source_invalid_output_type():

    data = {
        'source': 'mike',
        'test': 'true',
        'nested': {'foo': 'bar'}
    }

    output = 1
    field = Field(source='source', required=True)
    session = Session(field, data, output)
    with pytest.raises(FieldError):
        update_output_to_source(session)


def test_set_default():

    data = {
        'name': None,
        'test': 'true',
        'nested': {'foo': 'bar'}
    }

    output = {}

    field = Field(name='name', required=True, default='default')
    session = Session(field, data, output)
    session.data = data['name']
    set_default(session)
    assert session.data == 'default'


def test_set_default_falsy():
    # Default should not apply if value is falsy

    data = {
        'name': '',
        'test': 'true',
        'nested': {'foo': 'bar'}
    }

    output = {}

    field = Field(name='name', required=True, default='default')
    session = Session(field, data, output)
    session.data = data['name']
    set_default(session)
    assert session.data == ''



# kim/pipelines/datetime.py
# Copyright (C) 2014-2015 the Kim authors and contributors
# <see AUTHORS file>
#
# This module is part of Kim and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

import iso8601

from kim.utils import datetime as dt


from .base import pipe, is_valid_choice
from .marshaling import MarshalPipeline
from .serialization import SerializePipeline


def parse_iso8601(session):
    try:
        session.data = iso8601.parse_date(session.data)
    except iso8601.ParseError:
        raise session.field.invalid(error_type='invalid')

    return session.data


def parse_date_str(session):
    date_format = session.field.opts.date_format

    try:
        session.data = dt.strptime(session.data, date_format)
    except ValueError:
        raise session.field.invalid(error_type='invalid')

    return session.data


@pipe()
def is_valid_datetime(session):
    """Pipe used to determine if a value can be coerced to a datetime

    :param session: Kim pipeline session instance

    """

    if session.data is not None:
        date_format = session.field.opts.date_format
        if date_format == 'iso8601':
            return parse_iso8601(session)
        else:
            return parse_date_str(session)


@pipe()
def format_datetime(session):
    """Convert date or datetime object into formatted string representation.
    """
    if session.data is not None:
        date_format = session.field.opts.date_format
        if date_format == 'iso8601':
            session.data = session.data.isoformat()
        else:
            session.data = session.data.strftime(date_format)
    return session.data


class DateTimeMarshalPipeline(MarshalPipeline):
    """DateTimeMarshalPipeline

    .. seealso::
        :func:`kim.pipelines.datetime.is_valid_datetime`
        :func:`kim.pipelines.base.is_valid_choice`
        :class:`kim.pipelines.marshaling.MarshalPipeline`
    """

    validation_pipes = \
        [is_valid_datetime, is_valid_choice] + MarshalPipeline.validation_pipes


class DateTimeSerializePipeline(SerializePipeline):
    """DateTimeSerializePipeline

    .. seealso::
        :func:`kim.pipelines.datetime.format_datetime`
        :class:`kim.pipelines.marshaling.MarshalPipeline`
    """
    process_pipes = [format_datetime, ] + SerializePipeline.process_pipes


@pipe()
def cast_to_date(session):
    """cast session.data datetime object to a date() instance
    """
    if session.data is not None:
        session.data = session.data.date()
    return session.data


class DateMarshalPipeline(DateTimeMarshalPipeline):
    """DateMarshalPipeline

    .. seealso::
        :func:`kim.pipelines.datetime.cast_to_date`
        :class:`kim.pipelines.marshaling.MarshalPipeline`
    """

    process_pipes = DateTimeMarshalPipeline.process_pipes + [cast_to_date]


class DateSerializePipeline(DateTimeSerializePipeline):
    """DateSerializePipeline

    .. seealso::
        :func:`kim.pipelines.serialization.SerializePipeline`
    """
    pass



# kim/pipelines/__init__.py
# Copyright (C) 2014-2015 the Kim authors and contributors
# <see AUTHORS file>
#
# This module is part of Kim and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

from .base import *
from .string import *
from .numeric import *
from .nested import *
from .collection import *
from .boolean import *
from .static import *
from .datetime import *



# kim/pipelines/static.py
# Copyright (C) 2014-2015 the Kim authors and contributors
# <see AUTHORS file>
#
# This module is part of Kim and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

from .base import pipe
from .serialization import SerializePipeline


@pipe(run_if_none=True)
def get_static_value(session):
    """return the static value specified in FieldOpts
    """

    session.data = session.field.opts.value

    return session.data


class StaticSerializePipeline(SerializePipeline):
    """StaticSerializePipeline

    .. seealso::
        :func:`kim.pipelines.static.get_static_value`
        :class:`kim.pipelines.serialization.SerializePipeline`
    """

    process_pipes = [get_static_value, ] + SerializePipeline.process_pipes



# kim/roles.py
# Copyright (C) 2014-2015 the Kim authors and contributors
# <see AUTHORS file>
#
# This module is part of Kim and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

from kim.exception import RoleError


class Role(set):
    """Roles are a fundamental feature of Kim.  It's very common to need
    to provide a different view of your data or to only require a selection of
    fields when marshaling data.  ``Roles`` in Kim allow users
    to shape their data at runtime in a simple yet flexible manner.

    ``Roles`` are added to your :py:class:`~.Mapper` declarations
    using the ``__roles__`` attribute.

    Usage::

        from kim import Mapper, whitelist, field

        class UserMapper(Mapper):
            __type__ = User

            id = field.Integer(read_only=True)
            name = field.String(required=True)
            company = field.Nested('myapp.mappers.CompanyMapper')

            __roles__ = {
                'id_only': whitelist('id')
            }

    """

    def __init__(self, *args, **kwargs):
        """initialise a new :class:`Role`.

        :param whitelist:  pass a boolean indicating whether this
            role is a whitelist

        """
        self.whitelist = kwargs.pop('whitelist', True)
        super(Role, self).__init__(args)

    @property
    def fields(self):
        """return an iterable containing all the field names defined in this
        role.

        :rtype: list
        :returns: iterable of field names
        """

        return [k for k in self]

    def __contains__(self, field_name):
        """overloaded membership test that inverts the check depending on
        wether the role is a whitelist or blacklist.

        If the role is defined as whitelist=True the normal membership test
        is applied ie::

            >>> 'name' in whitelist('name')
            True

        For blacklist the test is flipped as we are aiming to ensure the field
        name is not present in the role::

            >>> 'other_name' in blacklist('name')
            True
            >>> 'name' in blacklist('name')
            False

        :param field_name: name of a field to test for membership
        :rtype: boolean
        :returns: boolean indicating wether field_name is found in the role
        """
        if self.whitelist:
            return super(Role, self).__contains__(field_name)
        else:
            return not super(Role, self).__contains__(field_name)

    def __or__(self, other):
        """Override handling of producing the union of two Roles to provide
        native support for merging whitelist and blacklist roles correctly.

        This overloading allows users to produce the union of two roles that
        may, on one side, want to allow fields and on the other exclude them.

        Usage::

            >>> from kim.role import whitelist, blacklist
            >>> my_role = whitelist('foo', 'bar') | blacklist('foo', 'baz')
            >>> my_role
            Role('bar')

        :param other: another instance of :class:`kim.role.Role`
        :raises: :class:`kim.exception.RoleError`
        :rtype: :class:`kim.role.Role`
        :returns: a new :class:`kim.role.Role` containng the set of field names

        """
        if not isinstance(other, Role):
            raise RoleError('union of built in types is not supported with roles')

        whitelist = True

        if self.whitelist and other.whitelist:
            # both roles are whitelists, return the union of both sets
            result = super(Role, self).__or__(other)

        elif self.whitelist and not other.whitelist:
            # we need to remove the fields in self(whitelist)
            # that appear in other(blacklist)
            result = super(Role, self).__sub__(other)

        elif not self.whitelist and other.whitelist:
            # Same as above, except we are keeping the fields from other
            result = other.__sub__(self)

        else:
            # both roles are blacklist, union them and set whitelist=False
            whitelist = False
            result = super(Role, self).__or__(other)

        return Role(*[k for k in result], whitelist=whitelist)

    def __and__(self, other):
        """Override handling of producing the intersection of two Roles to provide
        native support for merging whitelist and blacklist roles correctly.

        This overloading allows users to produce the intersection of two roles that
        may, on one side, want to allow fields and on the other exclude them.

        .. codeblock:: python

            >>>from kim.role import whitelist, blacklist
            >>>my_role = whitelist('foo', 'bar') & blacklist('foo', 'baz')
            >>>my_role
            Role('bar')

        :param other: another instance of :py:class:``.Role``
        :raises: :py:class:`.RoleError``

        :rtype: :py:class:``.Role``
        :returns: a new :py:class:``.Role`` containng the set of field names

        """
        if not isinstance(other, Role):
            raise RoleError('intersection of built types is '
                            'not supported with roles')

        whitelist = True

        if self.whitelist and other.whitelist:
            # both roles are whitelists, return the union of both sets
            result = super(Role, self).__and__(other)

        elif self.whitelist and not other.whitelist:
            # we need to remove the fields in self(whitelist)
            # that appear in other(blacklist)
            result = super(Role, self).__sub__(other)

        elif not self.whitelist and other.whitelist:
            # Same as above, except we are keeping the fields from other
            result = other.__sub__(self)

        else:  # both roles are blacklist, union them and set whitelist=False
            whitelist = False
            result = super(Role, self).__or__(other)

        return Role(*[k for k in result], whitelist=whitelist)


class whitelist(Role):
    """ Whitelists are roles that define a list of fields that are
    permitted for inclusion when marhsaling or serializing.
    For example, a whitelist role called ``id_only`` that contains
    the field name ``id`` instructs kim that whenever
    the ``id_only`` role is used **only** the ``id`` field should be
    considered in the input/output data.

    Usage::

        from kim import whitelist

        id_only_role = whitelist('id')

        class IdMixin(object):

            id = fields.Integer(read_only=True)

            __roles__ = {
                'id_only': id_only
            }
    """

    def __init__(self, *args, **kwargs):
        self.whitelist = True
        kwargs['whitelist'] = True
        super(whitelist, self).__init__(*args, **kwargs)


class blacklist(Role):
    """ Blacklists are role that act in the opposite manner to whitelists.
    They define a list of fields that should not be used
    when marshaling and serializing data.  A blacklist role named ``id_less``
    that contained the field name ``id`` would instruct kim that every
    field defined on the mapper should be considered except ``id``.

    Usage::

        from kim import whitelist

        class UserMapper(Mapper):

            id_less_role = blacklist('id')

            __roles__ = {
                'id_less': blacklist('id')
            }
    """

    def __init__(self, *args, **kwargs):
        kwargs['whitelist'] = False
        super(blacklist, self).__init__(*args, **kwargs)



# kim/pipelines/base.py
# Copyright (C) 2014-2015 the Kim authors and contributors
# <see AUTHORS file>
#
# This module is part of Kim and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

from itertools import chain
from functools import wraps

from kim.exception import StopPipelineExecution, FieldError
from kim.utils import attr_or_key, set_attr_or_key, attr_or_key_update, _remove_escapes


class Session(object):
    """Session objects acts as store for the state passed between
    one pipe method to another.

    Everytime a :class:`kim.field.Field` is marshaled or serialized a session object
    is created for each field that persists across every pipe inside of the fields
    MarsahlingPipeline or SerializationPipeline.

    Each pipe in a pipeline is able to work with the data set by the previous pipe using
    the session object.

    Example::

        from kim.pipelines.base import pipe

        @pipe()
        def odds_evens(session):

            if session.data is not None:
                if session.data % 2:
                    session.data = 'evens'
                else:
                    session.data = 'odds'

    In the example above our pipe changes the data of our pipeline based on the current
    value of session.data.

    As well as providing the interface to the data in a fields pipeline, the Session
    object also provides pipes with access to the overal state of that marshaling or
    serialization pipeline.
    """

    __slots__ = ('field', 'data', 'output', 'parent', 'mapper_session', 'nested_mapper')

    def __init__(self, field=None, data=None, output=None,
                 parent=None, mapper_session=None, nested_mapper=None):
        """Construct a new session.

        :param field: an instance of :class:`kim.field.Field` the scope of this session
            is bound too.
        :param data: data that has been passed along the pipeline field marshaling or
            serialization session.
        :param output: An object that contains the output of this fields marshaling
            or serialization session.
        :param parent:  If this is a wrapped field, then the parent kwarg will be a
            referrence to the instance of the field wrapping field.
        :param mapper_session: The overal mapper marshaling or serialization session
            this field session belongs to.
        """
        self.field = field
        self.data = data
        self.output = output
        self.parent = parent
        self.mapper_session = mapper_session
        self.nested_mapper = nested_mapper

    @property
    def mapper(self):
        """Return the :class:`kim.mapper.Mapper` bound to the scope of this Session.

        :returns: The :class:`kim.mapper.Mapper` bound to this Session.
        :rtype: :class:`kim.mapper.Mapper`
        """
        return self.mapper_session.mapper


def pipe(**pipe_kwargs):
    """Pipe decorator is provided as a convenience to avoid duplicating logic like
    not running pipes when session.data is null.

    :param run_if_none: Specify wether the pipe function should be called if session.data
        is None.

    Usage::

        from kim.pipelines.base import pipe

        @pipe(run_if_none=True)
        def my_pipe(session):

            do_stuff(session)
    """

    def pipe_decorator(pipe_func):

        @wraps(pipe_func)
        def inner(session, *args, **kwargs):

            if session.data is not None:
                return pipe_func(session)
            elif session.data is None and pipe_kwargs.get('run_if_none'):
                return pipe_func(session)
            else:
                return session.data

        return inner

    return pipe_decorator


class Pipeline(object):
    """Pipelines provide a simple, extensible way of processing data for
    a :class:`kim.field.Field`.  Each pipeline provides 4 input groups,
    ``input_pipes``, ``validation_pipes``, ``process_pipes`` and ``output_pipes``.
    Each containing `pipe` functions that are called in order
    passing data from one pipe to another.

    Kim pipes are similar to unix pipes, where each pipe in the chain has a
    single role in handling data before passing it on to the next pipe in the
    chain.

    Pipelines are typically ignorant to whether they
    are marhsaling data or serializing data, they simply take data in one end,
    this may be a deserialized dict of JSON or an object
    that's been populated from the database, and produce an output
    at the other.

    Usage::

        from kim.pipelines.base import Pipeline

        class StringIntPipeline(Pipeline):

            input_pipes = [get_data_from_json]
            validation_pipes = [is_numeric_string]
            process_pipes [cast_to_int]
            output_pipes = [update_output]
    """

    input_pipes = []
    validation_pipes = []
    process_pipes = []
    output_pipes = []

    __slots__ = ()

    @classmethod
    def get_pipeline(cls, **extra_pipes):
        chain = []
        chain.extend(cls.input_pipes + extra_pipes.get('input', []))
        chain.extend(cls.validation_pipes + extra_pipes.get('validation', []))
        chain.extend(cls.process_pipes + extra_pipes.get('process', []))
        chain.extend(cls.output_pipes + extra_pipes.get('output', []))

        return chain


def run_pipeline(pipeline, session, field, **opts):
    """ Iterate over all of the defined ``pipes`` for this pipeline.

    :param parent_session: The field being processed by this Pipeline is wrapped,
        parent_session will be passed and set on the fields session.
    :returns: Returns the output of the pipelines session.
    :rtype: mixed
    """

    # chain all the pipelines pipes together and process them until the all the
    # pipe groups have been exhausted or until
    # :class:`kim.exception.StopPipelineExecution` is raised.
    try:
        for pipe_func in pipeline:
            pipe_func(session)

        return session.output

    except StopPipelineExecution:
        return session.output


@pipe(run_if_none=True)
def get_data_from_name(session):
    """Extracts a specific key from data using ``field.name``.  This pipe is
    typically used as the entry point to a chain of input pipes.

    :param session: Kim pipeline session instance

    :rtype: mixed
    :returns: the key found in data using field.name

    """

    # If the field is wrapped by another field then the relevant data
    # will have already been pulled from the name.
    if session.field.opts._is_wrapped:
        return session.data

    value = attr_or_key(session.data, session.field.name)

    if value is None:
        if session.field.opts.required and session.field.opts.default is None:
            raise session.field.invalid(error_type='required')
        elif session.field.opts.default is not None:
            session.data = session.field.opts.default
            return session.data
        elif not session.field.opts.allow_none:
            raise session.field.invalid(error_type='none_not_allowed')

    session.data = value
    return session.data


@pipe()
def get_data_from_source(session):
    """Extracts a specific key from data using ``field.source``.  This pipe is
    typically used as the entry point to a chain of output pipes.

    :param session: Kim pipeline session instance

    :rtype: mixed
    :returns: the key found in data using field.source

    """

    source = session.field.opts.source

    # If the field is wrapped by another field then the relevant data
    # will have already been pulled from the source.
    if session.field.opts._is_wrapped or source == '__self__':
        return session.data

    value = attr_or_key(session.data, source)
    session.data = value
    return session.data


@pipe(run_if_none=True)
def get_field_if_required(session):

    if session.data is None:
        session.data = session.field.opts.default

    return session.data


@pipe()
def read_only(session):
    """End processing of a pipeline if a Field is marked as read_only.

    :param session: Kim pipeline session instance

    :raises  StopPipelineExecution:
    """

    if session.field.opts.read_only:
        raise StopPipelineExecution('read_only field')

    return session.data


@pipe()
def is_valid_choice(session):
    """End processing of a pipeline if a Field is marked as read_only.

    :param session: Kim pipeline session instance

    :raises  StopPipelineExecution:
    """

    choices = session.field.opts.choices
    if choices is not None and session.data not in choices:
        raise session.field.invalid('invalid_choice')

    return session.data


@pipe(run_if_none=True)
def update_output_to_name(session):
    """Store ``data`` at ``field[name]`` for a ``field`` inside
    of ``output``

    :param session: Kim pipeline session instance

    :returns: None
    """
    session.output[_remove_escapes(session.field.name)] = session.data


@pipe(run_if_none=True)
def update_output_to_source(session):
    """Store ``data`` at field.opts.source for a ``field`` inside
    of ``output``

    :param session: Kim pipeline session instance

    :raises: FieldError
    :returns: None
    """

    source = _remove_escapes(session.field.opts.source)
    try:
        if source == '__self__':
            attr_or_key_update(session.output, session.data)
        else:
            set_attr_or_key(session.output, session.field.opts.source, session.data)
    except (TypeError, AttributeError):
        raise FieldError('output does not support attribute or '
                         'key based set operations')


@pipe(run_if_none=True)
def set_default(session):
    """If ``data`` is None, set default if it is set.

    :param session: Kim pipeline session instance

    :returns: None
    """
    if session.data is None:
        session.data = session.field.opts.default

    return session.data



# kim/pipelines/numeric.py
# Copyright (C) 2014-2015 the Kim authors and contributors
# <see AUTHORS file>
#
# This module is part of Kim and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

from decimal import Decimal, InvalidOperation

from .base import pipe, is_valid_choice
from .marshaling import MarshalPipeline
from .serialization import SerializePipeline


@pipe()
def is_valid_integer(session):
    """Pipe used to determine if a value can be coerced to an int

    :param session: Kim pipeline session instance

    """

    try:
        session.data = int(session.data)
    except TypeError:
        raise session.field.invalid(error_type='type_error')
    except ValueError:
        raise session.field.invalid(error_type='type_error')
    return session.data


@pipe()
def bounds_check(session):
    """Pipe used to determine if a value is within the min and max bounds on
    the field

    :param session: Kim pipeline session instance

    """

    max_ = session.field.opts.max
    min_ = session.field.opts.min

    if max_ is not None and session.data > max_:
        raise session.field.invalid(error_type='out_of_bounds')
    if min_ is not None and session.data < min_:
        raise session.field.invalid(error_type='out_of_bounds')

    return session.data


class IntegerMarshalPipeline(MarshalPipeline):
    """IntegerMarshalPipeline

    .. seealso::
        :func:`kim.pipelines.numeric.is_valid_integer`
        :func:`kim.pipelines.base.is_valid_choice`
        :func:`kim.pipelines.numeric.bounds_check`
        :class:`kim.pipelines.marshaling.MarshalPipeline`
    """

    validation_pipes = \
        [is_valid_integer, is_valid_choice, bounds_check] + MarshalPipeline.validation_pipes


class IntegerSerializePipeline(SerializePipeline):
    """IntegerSerializePipeline

    .. seealso::
        :class:`kim.pipelines.serialization.SerializePipeline`
    """
    pass


@pipe()
def is_valid_decimal(session):
    """Pipe used to determine if a value can be coerced to a Decimal

    :param session: Kim pipeline session instance

    """
    try:
        return Decimal(session.data)
    except InvalidOperation:
        raise session.field.invalid(error_type='type_error')


@pipe()
def coerce_to_decimal(session):
    """Coerce str representation of a decimal into a valid Decimal object.
    """
    decimals = session.field.opts.precision
    precision = Decimal('0.' + '0' * (decimals - 1) + '1')
    session.data = Decimal(session.data).quantize(precision)
    return session.data


class DecimalMarshalPipeline(MarshalPipeline):
    """DecimalMarshalPipeline

    .. seealso::
        :func:`kim.pipelines.numeric.is_valid_decimal`
        :func:`kim.pipelines.numeric.coerce_to_decimal`
        :class:`kim.pipelines.marshaling.MarshalPipeline`
    """

    validation_pipes = [is_valid_decimal, coerce_to_decimal, bounds_check] + MarshalPipeline.validation_pipes


# TODO(mike) This should probably move to base
@pipe()
def to_string(session):
    """coerce decimal value into str so it's valid for json
    """
    if session.data is not None:
        session.data = str(session.data)
        return session.data


class DecimalSerializePipeline(SerializePipeline):
    """DecimalSerializePipeline

    .. seealso::
        :func:`kim.pipelines.numeric.coerce_to_decimal`
        :func:`kim.pipelines.numeric.to_string`
        :class:`kim.pipelines.serialization.SerializePipeline`
    """

    process_pipes = [coerce_to_decimal, to_string] + SerializePipeline.process_pipes


@pipe()
def is_valid_float(session):
    """Pipe used to determine if a value can be coerced to a Float

    :param session: Kim pipeline session instance

    """
    try:
        return float(session.data)
    except (InvalidOperation, ValueError):
        raise session.field.invalid(error_type='type_error')



@pipe()
def coerce_to_float(session):
    """Coerce str representation of a decimal into a valid Float object.
    """

    decimals = session.field.opts.precision
    session.data = round(float(session.data), decimals)
    return session.data


class FloatMarshalPipeline(MarshalPipeline):
    """FloatMarshalPipeline

    .. seealso::
        :func:`kim.pipelines.numeric.is_valid_decimal`
        :func:`kim.pipelines.numeric.coerce_to_decimal`
        :class:`kim.pipelines.marshaling.MarshalPipeline`
    """

    validation_pipes = [is_valid_float, bounds_check] + MarshalPipeline.validation_pipes
    process_pipes = [coerce_to_float] + MarshalPipeline.process_pipes


class FloatSerializePipeline(SerializePipeline):
    """FloatSerializePipeline

    .. seealso::
        :func:`kim.pipelines.numeric.coerce_to_float`
        :func:`kim.pipelines.numeric.to_string`
        :class:`kim.pipelines.serialization.SerializePipeline`
    """

    process_pipes = [coerce_to_float, to_string] + SerializePipeline.process_pipes



# kim/pipelines/nested.py
# Copyright (C) 2014-2015 the Kim authors and contributors
# <see AUTHORS file>
#
# This module is part of Kim and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

from kim.utils import attr_or_key

from .base import pipe
from .marshaling import MarshalPipeline
from .serialization import SerializePipeline


def _call_getter(session):
    if session.field.opts.getter:
        result = session.field.opts.getter(session)
        return result


@pipe()
def marshal_nested(session):
    """Marshal data using the nested mapper defined on this field.

    There are 6 possible scenarios, depending on the security setters and
    presence of a getter function

    * Getter function returns an object and no updates are allowed - Return the
      object immediately
    * Getter function returns an object and updates are allowed - Call the
      nested mapper with the object to update it
    * Object already exists, getter function returns None/does not exist and
      in place updates are allowed - Call the nested mapper with the existing
      object to update it
    * Getter function returns None/does not exist and creation of new objects
      is allowed - Call the nested mapper to create a new object
    * Getter function returns None/does not exist and creation of new objects
      is not allowed, nor are in place updates - Raise an exception.
    * Object already exists, getter function returns None/does not exist and
      partial updates are allowed - Call the nested mapper with the existing object
      to update it

    :param session: Kim pipeline session instance
    """

    resolved = _call_getter(session)

    partial = session.mapper_session.partial
    parent_mapper = session.mapper

    if session.parent and session.parent.nested_mapper:
        nested_mapper_class = session.parent.nested_mapper
    else:
        nested_mapper_class = session.field.get_mapper(as_class=True)

    if resolved is not None:
        if session.field.opts.allow_updates:
            nested_mapper = nested_mapper_class(
                data=session.data, obj=resolved, partial=partial,
                parent=parent_mapper)
            session.data = nested_mapper.marshal(role=session.field.opts.role)
        else:
            session.data = resolved
    else:
        existing_value = attr_or_key(session.output, session.field.name)
        if (session.field.opts.allow_updates_in_place or
                session.field.opts.allow_partial_updates) and \
                existing_value is not None:
            nested_mapper = nested_mapper_class(
                data=session.data, obj=existing_value, partial=partial,
                parent=parent_mapper)
            session.data = nested_mapper.marshal(role=session.field.opts.role)
        elif session.field.opts.allow_create:
            nested_mapper = nested_mapper_class(
                data=session.data, partial=partial, parent=parent_mapper)
            session.data = nested_mapper.marshal(role=session.field.opts.role)
        else:
            raise session.field.invalid(error_type='not_found')

    return session.data


@pipe(run_if_none=True)
def serialize_nested(session):
    """Serialize data using the nested mapper defined on this field.

    :param session: Kim pipeline session instance
    """

    if session.data is None:
        session.data = session.field.opts.null_default
        return session.data

    # Grab the Mapper defined for the nested field and call serialize()
    if session.parent and session.parent.nested_mapper:
        nested_mapper = session.parent.nested_mapper(obj=session.data)
    else:
        nested_mapper = session.field.get_mapper(obj=session.data)

    session.data = nested_mapper.serialize(role=session.field.opts.role)

    return session.data


class NestedMarshalPipeline(MarshalPipeline):
    """NestedMarshalPipeline

    .. seealso::
        :func:`kim.pipelines.nested.marshal_nested`
        :class:`kim.pipelines.marshaling.MarshalPipeline`
    """

    output_pipes = [marshal_nested, ] + MarshalPipeline.output_pipes


class NestedSerializePipeline(SerializePipeline):
    """NestedSerializePipeline

    .. seealso::
        :func:`kim.pipelines.nested.serialize_nested`
        :class:`kim.pipelines.serialization.SerializePipeline`
    """

    process_pipes = [serialize_nested, ] + SerializePipeline.process_pipes



from kim.field import Static
from kim.pipelines.base import Session
from kim.pipelines.static import get_static_value

from ..conftest import get_mapper_session


def test_get_static_value():

    field = Static('foo', name='test')
    session = Session(field, 'bar', {})

    assert get_static_value(session) == 'foo'


def test_static_read_only():

    field = Static('foo', name='test')
    assert field.opts.read_only

    field = Static('foo', read_only=False, name='test')
    assert field.opts.read_only


def test_static_output():

    class Foo(object):
        name = 'value'

    field = Static('foo', name='name', required=True)

    output = {}
    mapper_session = get_mapper_session(obj=Foo(), output=output)
    field.serialize(mapper_session)
    assert output == {'name': 'foo'}
